

import ngsolve as ng
from netgen import csg, occ
from netgen import meshing


#%%  Functions

def MakeMesh(sphere_radius):
    """
    Create space mesh
    """
    prism = csg.OrthoBrick( csg.Pnt(-2,-2,-1), csg.Pnt(2,2,1) )
    sphere = csg.Sphere( csg.Pnt(0,0,0), sphere_radius)

    # generate sphere and mesh it
    geo1 = csg.CSGeometry()
    geo1.Add(sphere)
    m1 = geo1.GenerateMesh()#maxh=0.1)
    m1.SetMaterial(1, 'vacuum')
    m1.Refine()

    # generate object from step file and mesh it
    geo2 = csg.CSGeometry()
    geo2.Add(prism)
    m2 = geo2.GenerateMesh()#(maxh=0.05)
    m2.SetMaterial(2, 'object')
    m2.Refine()

    # create an empty mesh
    mesh = meshing.Mesh()
    # define domaines
    fd_outside = mesh.Add (meshing.FaceDescriptor(bc=1, domin=1, surfnr=1))
    fd_inside = mesh.Add (meshing.FaceDescriptor(bc=2, domin=2, domout=1, surfnr=2))
    
    # copy all boundary points from first mesh to new mesh
    pmap1 = { }
    for e in m1.Elements2D():
        for v in e.vertices:
            if (v not in pmap1):
                pmap1[v] = mesh.Add (m1[v])
    # copy surface elements from first mesh to new mesh
    for e in m1.Elements2D():
        mesh.Add (meshing.Element2D (fd_outside, [pmap1[v] for v in e.vertices]))

    # same for the second mesh:
    pmap2 = { }
    for e in m2.Elements2D():
        for v in e.vertices:
            if (v not in pmap2):
                pmap2[v] = mesh.Add (m2[v])
    for e in m2.Elements2D():
        mesh.Add (meshing.Element2D (fd_inside, [pmap2[v] for v in e.vertices]))
    
    
    mesh.GenerateVolumeMesh()

    # Create ngsolve_mesh from ntegen_mesh
    ngmesh = ng.Mesh(mesh)

    return ngmesh, m2


def MakeFEspace(space_mesh, object_mesh, dRho):
    """
    Create Finate Element Space, set up equations and boundary conditions
    """

    density = ng.CoefficientFunction([0, dRho])

    # Create space
    fes = ng.H1(space_mesh, order=2)#, dirichlet=[1,2])
    u = fes.TrialFunction()  # symbolic object
    v = fes.TestFunction()   # symbolic object
    gfu = ng.GridFunction(fes)  # solution
    
    a = ng.BilinearForm(fes, symmetric=True)
    a += ng.grad(u)*ng.grad(v)*ng.dx

    f = ng.LinearForm(fes)
    f += density*v*ng.dx

    # H(div) space for flux recovery
    space_flux = ng.HDiv(space_mesh, order=2)
    gf_flux = ng.GridFunction(space_flux, "flux")
    flux = ng.grad(gfu)
    gf_flux.Set(flux)

    return fes, gfu, a, f, space_flux, gf_flux


def SolveBVP(fes, gfu, a, f):
    """
    Assemble equations and solve Boundary Value Problem
    """
    fes.Update()
    gfu.Update()
    a.Assemble()
    f.Assemble()
    gfu.vec.data = a.mat.Inverse(freedofs=fes.FreeDofs()) * f.vec
    return fes, gfu, a, f


def CalcError(space_mesh, gfu, space_flux, gf_flux):
    """
    gradient-recovery-type error estimator.
    Use the flux of the computed solution and interpolate it into the H(div) space.
    """
    # compute the flux:
    space_flux.Update()
    gf_flux.Update()
    flux = ng.grad(gfu)
    gf_flux.Set(flux)

    # compute estimator:
    err = (flux-gf_flux)*(flux-gf_flux)
    eta2 = ng.Integrate(err, space_mesh, ng.VOL, element_wise=True)
    maxerr = max(eta2)

    # Mark elements with large error estimator for refinement:
    for el in space_mesh.Elements(ng.BND):
        space_mesh.SetRefinementFlag(el, False)
    for el in space_mesh.Elements():
        space_mesh.SetRefinementFlag(el, eta2[el.nr] > 0.25*maxerr)

    return space_mesh, space_flux, gf_flux, maxerr


def AdaptiveRefine(err_threshold, space_mesh, fes, gfu, a, f, space_flux, gf_flux):
    """
    Adaptive loop refinement
    """
    fes, gfu, a, f = SolveBVP(fes, gfu, a, f)
    maxerr = 1
    while maxerr>err_threshold:
        space_mesh, space_flux, gf_flux, maxerr = CalcError(space_mesh, gfu, space_flux, gf_flux)
        print("ndof =", fes.ndof, " maxerr =", maxerr)
        space_mesh.Refine()
        fes, gfu, a, f = SolveBVP(fes, gfu, a, f)
    return gfu, space_mesh


def SolveAdaptive(space_mesh, object_mesh, dRho, err_threshold):
    fes, gfu, a, f, space_flux, gf_flux = MakeFEspace(space_mesh, object_mesh, dRho)
    gfu, space_mesh = AdaptiveRefine(err_threshold, space_mesh, fes, gfu, a, f, space_flux, gf_flux)
    return gfu, space_mesh


#%% Calculation

dRho = 1
sphere_radius = 100
err_threshold = 1e-2


space_mesh, object_mesh = MakeMesh(sphere_radius)
gfu, space_mesh = SolveAdaptive(space_mesh, object_mesh, dRho, err_threshold)

